@@ -17,11 +17,13 @@
 
 """Command line interface Thamos for interaction with Thoth."""
 
-import logging
-import typing
-import os
-import sys
 import json
+import logging
+import os
+import shutil
+import subprocess
+import sys
+import typing
 from functools import wraps
 from typing import Tuple
 from typing import Optional
@@ -35,6 +37,7 @@ from rich.text import Text
 from rich.table import Table
 from rich import box
 import daiquiri
+from micropipenv import HashMismatch
 from thoth.python import Project
 from thoth.common import cwd
 from thoth.common import ThothAdviserIntegrationEnum
@@ -44,6 +47,9 @@ from thamos.exceptions import NoRuntimeEnvironmentError
 from thamos.config import config as configuration
 from thamos.lib import advise_here as thoth_advise_here
 from thamos.lib import provenance_check as thoth_provenance_check
+from thamos.lib import list_hardware_environments
+from thamos.lib import list_python_package_indexes
+from thamos.lib import list_thoth_s2i
 from thamos.lib import get_log
 from thamos.lib import install as thamos_install
 from thamos.lib import get_status
@@ -138,7 +144,7 @@ def _write_files(
     project = Project.from_dict(requirements, requirements_lock)
     if requirements_format == "pipenv":
         _LOGGER.debug("Writing to Pipfile/Pipfile.lock in %r", os.getcwd())
-        project.to_files()
+        project.to_files(keep_thoth_section=True)
     elif requirements_format in ("pip", "pip-tools", "pip-compile"):
         _LOGGER.debug("Writing to requirements.in/requirements.txt in %r", os.getcwd())
         project.to_pip_compile_files()
@@ -295,7 +301,12 @@ class AliasedGroup(click.Group):
     default=None,
     help="Use selected host instead of the one stated in the configuration file.",
 )
-def cli(ctx=None, verbose: bool = False, workdir: str = None, thoth_host: str = None):
+def cli(
+    ctx=None,
+    verbose: bool = False,
+    workdir: str = typing.Optional[None],
+    thoth_host: str = None,
+):
     """CLI tool for interacting with Thoth."""
     if ctx:
         ctx.auto_envvar_prefix = "THAMOS"
@@ -360,6 +371,7 @@ def _print_version(ctx, json_output: bool = False):
 
 
 @cli.command("install")
+@click.pass_context
 @click.option(
     "--dev/--no-dev",
     is_flag=True,
@@ -377,16 +389,160 @@ def _print_version(ctx, json_output: bool = False):
     envvar="THAMOS_RUNTIME_ENVIRONMENT",
     help="Specify explicitly runtime environment for which the installation process should be done.",
 )
-def install(runtime_environment: str, dev: bool) -> None:
-    """Install dependencies as stated in Pipfile.lock or requirements.txt.
+@click.argument("pip_args", nargs=-1)
+@handle_cli_exception
+def install(runtime_environment: str, dev: bool, pip_args: Tuple[str]) -> None:
+    """Install dependencies as stated in the lock file.
 
     This command assumes requirements files are present and dependencies are already resolved.
-    If that's not the case, issue `thamos advise` before calling this.
+    If that's not the case, issue `thamos advise` before running this command.
+
+    Examples:
+      thamos install --runtime-environment "testing"
+
+      thamos install --dev
+
+      thamos install --no-dev -- --force-reinstall
     """
-    thamos_install(runtime_environment_name=runtime_environment, dev=dev)
+    try:
+        thamos_install(
+            runtime_environment_name=runtime_environment, dev=dev, pip_args=pip_args
+        )
+    except HashMismatch as exc:
+        _LOGGER.error(
+            "Changes made to packages are not reflected in the lock file, please install "
+            "using `thamos advise --install` to apply changes made: %s",
+            str(exc),
+        )
+
+
+def _error_virtual_environment(virtualenv_path: str) -> None:
+    """Print error and exit if virtual environment was not setup."""
+    if not os.path.isdir(virtualenv_path):
+        if configuration.content.get("virtualenv", False):
+            _LOGGER.error(
+                "No virtual environment created yet, create it using `thamos install`"
+            )
+        else:
+            _LOGGER.error(
+                "No virtual environment management configured, use `thamos config` to adjust configuration"
+            )
+
+        sys.exit(1)
+
+
+@cli.command("run")
+@click.argument("command", nargs=-1, metavar="CMD")
+@click.option(
+    "--runtime-environment",
+    "-r",
+    type=str,
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify explicitly runtime environment to get recommendations for; "
+    "defaults to the first entry in the configuration file.",
+)
+def run(command: typing.List[str], runtime_environment: Optional[str] = None) -> None:
+    """Run the command in virtual environment.
+
+    Examples:
+      thamos run ./app.py
+
+      thamos run --runtime-environment "testing" -- flask --help
+    """
+    virtualenv_path = configuration.get_virtualenv_path(runtime_environment)
+    if virtualenv_path is None:
+        _LOGGER.error("No virtual environment found for %r", runtime_environment)
+        sys.exit(1)
+
+    python_path = os.path.join(virtualenv_path, "bin", "python3")
+    _error_virtual_environment(virtualenv_path)
+    # No magic here.
+    cmd = [python_path, *command]
+    subprocess.run(cmd, universal_newlines=True, check=True)
+
+
+@cli.command("venv")
+@click.pass_context
+@click.option(
+    "--runtime-environment",
+    "-r",
+    type=str,
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify explicitly runtime environment to get recommendations for; "
+    "defaults to the first entry in the configuration file.",
+)
+def venv(ctx, runtime_environment: Optional[str] = None) -> None:
+    """Get path of the virtual environment."""
+    virtualenv_path = configuration.get_virtualenv_path(runtime_environment)
+    if virtualenv_path is None:
+        _LOGGER.error("No virtual environment found")
+        ctx.exit(1)
+    _error_virtual_environment(virtualenv_path)
+    print(virtualenv_path)
+
+
+@cli.command("purge")
+@click.pass_context
+@click.option(
+    "--runtime-environment",
+    "-r",
+    type=str,
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify explicitly runtime environment to get recommendations for; "
+    "defaults to the first entry in the configuration file.",
+)
+@click.option(
+    "--all",
+    "-A",
+    is_flag=True,
+    help="Purge virtual environments for all the runtime environments configured.",
+)
+def purge(ctx, runtime_environment: Optional[str] = None, all: bool = False) -> None:
+    """Remove virtual environment created.
+
+    Examples:
+      thamos purge
+
+      thamos purge --runtime-environment "testing"
+
+      thamos purge --all
+    """
+    if all:
+        for runtime_environment_config in configuration.list_runtime_environments():
+            _LOGGER.warning(
+                "Removing virtual environment for %r",
+                runtime_environment_config["name"],
+            )
+            path = configuration.get_virtualenv_path(runtime_environment),
+            if path is None:
+                _LOGGER.warning("No virtual environment for %r found", runtime_environment_config["name"])
+                continue
+            shutil.rmtree(
+                path,
+                ignore_errors=True,
+            )
+    else:
+        runtime_environment_config = configuration.get_runtime_environment(
+            runtime_environment
+        )
+        _LOGGER.warning(
+            "Removing virtual environment for %r", runtime_environment_config["name"]
+        )
+        path = configuration.get_virtualenv_path(runtime_environment),
+        if path is None:
+            _LOGGER.error("No virtual environment for %r found", runtime_environment_config["name"])
+            ctx.exit(1)
+        shutil.rmtree(path, ignore_errors=True)
 
 
 @cli.command("advise")
+@click.pass_context
 @click.option(
     "--debug",
     is_flag=True,
@@ -469,11 +625,12 @@ def install(runtime_environment: str, dev: bool) -> None:
     show_default=True,
     help="Write advised manifest changes to a file.",
 )
+@handle_cli_exception
 def advise(
     debug: bool = False,
     no_write: bool = False,
-    recommendation_type: str = None,
-    runtime_environment: str = None,
+    recommendation_type: typing.Optional[str] = None,
+    runtime_environment: typing.Optional[str] = None,
     no_wait: bool = False,
     no_static_analysis: bool = False,
     json_output: bool = False,
@@ -483,7 +640,22 @@ def advise(
     install: bool = False,
     write_advised_manifest_changes: Optional[str] = None,
 ):
-    """Ask Thoth for recommendations on application stack."""
+    """Ask Thoth for recommendations on the application stack.
+
+    Ask the remote Thoth service for advise on the application stack used. The command
+    collects information stated in the .thoth.yaml file for the given runtime environment,
+    static source code analysis and requirements for the application and sends them to the
+    remote service. Optionally, install packages resolved by Thoth.
+
+    Examples:
+      thamos advise --runtime-environment "testing"
+
+      thamos advise --dev
+
+      thamos advise --install
+
+      thamos advise --no-static-analysis --no-user-stack
+    """
     if install and no_wait:
         _LOGGER.error("Cannot install dependencies as --no-wait was provided")
         sys.exit(1)
@@ -501,6 +673,7 @@ def advise(
         # In CLI we always call to obtain only the best software stack (count is implicitly set to 1).
         results = thoth_advise_here(
             recommendation_type=recommendation_type,
+            src_path=configuration.config_path,
             runtime_environment_name=runtime_environment,
             debug=debug,
             nowait=no_wait,
@@ -589,6 +762,7 @@ def advise(
 
 
 @cli.command("provenance-check")
+@click.pass_context
 @click.option(
     "--debug",
     is_flag=True,
@@ -619,7 +793,6 @@ def advise(
     envvar="THAMOS_FORCE",
     help="Force analysis run bypassing server-side cache.",
 )
-@click.pass_context
 @handle_cli_exception
 def provenance_check(
     debug: bool = False,
@@ -628,7 +801,14 @@ def provenance_check(
     force: bool = False,
     runtime_environment: typing.Optional[str] = None,
 ):
-    """Check provenance of installed packages."""
+    """Check provenance of installed packages.
+
+    Collect information about direct dependencies and dependencies stated in the lock file
+    and send them to the remote service to verify their provenance.
+
+    Examples:
+      thamos provenance-check --runtime-environment "production"
+    """
     with cwd(configuration.get_overlays_directory(runtime_environment)):
         if configuration.requirements_format != "pipenv":
             raise ValueError(
@@ -669,11 +849,13 @@ def provenance_check(
 
 
 @cli.command("log")
-@click.argument("analysis_id", type=str, required=False)
-def log(analysis_id: str = None):
+@click.pass_context
+@click.argument("analysis_id", type=str, required=False, metavar="ANALYSIS_ID")
+@handle_cli_exception
+def log(analysis_id: typing.Optional[str] = None):
     """Get log of running or finished analysis.
 
-    If ANALYSIS_ID is not provided, there will be used last analysis id, if noted by Thamos.
+    If ANALYSIS_ID is not provided, the last request is used by default.
     """
     if not analysis_id:
         with workdir():
@@ -685,7 +867,8 @@ def log(analysis_id: str = None):
 
 
 @cli.command("status")
-@click.argument("analysis_id", type=str, required=False)
+@click.pass_context
+@click.argument("analysis_id", type=str, required=False, metavar="ANALYSIS_ID")
 @click.option(
     "--output-format",
     "-o",
@@ -693,10 +876,18 @@ def log(analysis_id: str = None):
     default="table",
     help="Specify output format for the status report.",
 )
-def status(analysis_id: str = None, output_format: str = None):
+@handle_cli_exception
+def status(
+    analysis_id: typing.Optional[str] = None, output_format: typing.Optional[str] = None
+):
     """Get status of an analysis.
 
-    If ANALYSIS_ID is not provided, there will be used last analysis id, if noted by Thamos.
+    If ANALYSIS_ID is not provided, the last request is used by default.
+
+    Examples:
+      thamos status
+
+      thamos status "adviser-940101080006-110c392feb7cf6da"
     """
     if not analysis_id:
         with workdir():
@@ -730,8 +921,14 @@ def status(analysis_id: str = None, output_format: str = None):
 
 
 @cli.command("list")
+@click.pass_context
+@handle_cli_exception
 def list_() -> None:
-    """List available runtime environments."""
+    """List available runtime environments configured.
+
+    Examples:
+      thamos list
+    """
     with workdir(configuration.CONFIG_NAME):
         environments = configuration.list_runtime_environments()
 
@@ -744,6 +941,7 @@ def list_() -> None:
 
 
 @cli.command("show")
+@click.pass_context
 @click.option(
     "--output-format",
     "-o",
@@ -761,8 +959,13 @@ def list_() -> None:
     envvar="THAMOS_RUNTIME_ENVIRONMENT",
     help="Specify explicitly runtime environment to be shown.",
 )
+@handle_cli_exception
 def show(output_format: str, runtime_environment: Optional[str] = None) -> None:
-    """Show configuration of available runtime environments."""
+    """Show details for configured runtime environments.
+
+    Examples:
+      thamos show --runtime-environment "development"
+    """
     with workdir(configuration.CONFIG_NAME):
         environments = configuration.list_runtime_environments()
 
@@ -793,6 +996,7 @@ def show(output_format: str, runtime_environment: Optional[str] = None) -> None:
 
 
 @cli.command("config")
+@click.pass_context
 @click.option(
     "--no-interactive",
     "-I",
@@ -808,11 +1012,13 @@ def show(output_format: str, runtime_environment: Optional[str] = None) -> None:
     envvar="THAMOS_CONFIG_TEMPLATE",
     help="Template which should be used instead of the default one.",
 )
+@handle_cli_exception
 def config(no_interactive: bool = False, template: str = None):
-    """Adjust Thamos and Thoth remote configuration.
+    """Adjust Thamos and Thoth configuration.
 
-    Perform autodiscovery of available hardware and software on the host and
-    create a default configuration for Thoth (placed into .thoth.yaml).
+    Open the .thoth.yaml configuration file. If does not exist yet, perform
+    auto-discovery of available hardware and software on the host and
+    create a default configuration for Thoth.
     """
     if template:
         _LOGGER.info(
@@ -841,6 +1047,7 @@ def config(no_interactive: bool = False, template: str = None):
 
 
 @cli.command("check")
+@click.pass_context
 @click.option(
     "--runtime-environment",
     "-r",
@@ -857,8 +1064,18 @@ def config(no_interactive: bool = False, template: str = None):
     default="table",
     help="Specify output format for the status report.",
 )
+@handle_cli_exception
 def check(runtime_environment: Optional[str], output_format: str) -> None:
-    """Check configuration file and runtime environment."""
+    """Check the configuration file and runtime environment.
+
+    Check the correctness of the configuration file and runtime environment configuration
+    for the current host.
+
+    Examples:
+        thamos check --runtime-environment "production"
+
+        thamos check --output-format yaml
+    """
     result = configuration.check(runtime_environment_name=runtime_environment)
 
     if output_format == "yaml":
@@ -901,4 +1118,298 @@ def check(runtime_environment: Optional[str], output_format: str) -> None:
     sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
 
 
+@cli.command("s2i")
+@click.pass_context
+@click.option(
+    "--output-format",
+    "-o",
+    type=click.Choice(["json", "yaml", "table"]),
+    default="table",
+    help="Specify output format for the status report.",
+)
+@handle_cli_exception
+def s2i(output_format: str) -> None:
+    """Check available Thoth Source-To-Image containers.
+
+    Examples:
+      thamos s2i --output-format json
+    """
+    result = list_thoth_s2i()
+
+    if output_format == "yaml":
+        yaml.safe_dump({"s2i": result}, sys.stdout)
+    elif output_format == "json":
+        json.dump({"s2i": result}, sys.stdout, indent=2)
+        sys.stdout.write("\n")
+    elif output_format == "table":
+        table = Table()
+
+        header = set()
+        for item in result:
+            for key in item.keys():
+                header.add(key)
+
+        header_sorted = sorted(header)
+        header_sorted.append("info")
+        for item in header_sorted:
+            table.add_column(
+                item.replace("_", " ").capitalize(),
+                style="cyan",
+                overflow="fold",
+            )
+
+        for item in result:
+            row = []
+            for key in header_sorted:
+                if key == "info":
+                    image_name = item.get("thoth_s2i_image_name")
+                    row.append(jl(image_name.rsplit("/", maxsplit=1)[-1]))
+                    continue
+
+                entry = item.get(key)
+                row.append(str(entry) if entry is not None else "-")
+
+            table.add_row(*row)
+
+        console = Console()
+        console.print(table, justify="center")
+
+    sys.exit(1 if any(item.get("type") == "ERROR" for item in result) else 0)
+
+
+@cli.command("hw")
+@click.pass_context
+@click.option(
+    "--output-format",
+    "-o",
+    type=click.Choice(["json", "yaml", "table"]),
+    default="table",
+    help="Specify output format for the status report.",
+)
+@handle_cli_exception
+def hw(output_format: str) -> None:
+    """List available hardware information on backend.
+
+    List available hardware information on the backend for which resolver can give more detailed information.
+    This hardware configuration is not enforced and is find if it does not match the one available on the
+    client side.
+    """
+    result = list_hardware_environments()
+
+    if output_format == "yaml":
+        yaml.safe_dump({"hw": result}, sys.stdout)
+    elif output_format == "json":
+        json.dump({"hw": result}, sys.stdout, indent=2)
+        sys.stdout.write("\n")
+    elif output_format == "table":
+        table = Table()
+
+        header = set()
+        for item in result:
+            for key in item.keys():
+                header.add(key)
+
+        header_sorted = sorted(header)
+        for item in header_sorted:
+            table.add_column(
+                item.replace("_", " ")
+                .replace("cpu", "CPU")
+                .replace("gpu", "GPU")
+                .replace("ram", "RAM"),
+                style="cyan",
+                overflow="fold",
+            )
+
+        for item in result:
+            row = []
+            for key in header_sorted:
+                entry = item.get(key)
+                row.append(str(entry) if entry is not None else "-")
+
+            table.add_row(*row)
+
+        console = Console()
+        console.print(table, justify="center")
+
+    sys.exit(0)
+
+
+@cli.command("indexes")
+@click.pass_context
+@click.option(
+    "--output-format",
+    "-o",
+    type=click.Choice(["json", "yaml", "table"]),
+    default="table",
+    help="Specify output format for the status report.",
+)
+@handle_cli_exception
+def indexes(output_format: str) -> None:
+    """List available Python package indexes."""
+    result = list_python_package_indexes()
+
+    if output_format == "yaml":
+        yaml.safe_dump({"indexes": result}, sys.stdout)
+    elif output_format == "json":
+        json.dump({"indexes": result}, sys.stdout, indent=2)
+        sys.stdout.write("\n")
+    elif output_format == "table":
+        table = Table()
+
+        header = set()
+        for item in result:
+            for key in item.keys():
+                if key == "warehouse_api_url":
+                    continue
+                header.add(key)
+
+        header_sorted = sorted(header)
+        for item in header_sorted:
+            table.add_column(
+                item.capitalize()
+                .replace("_", " ")
+                .replace("Url", "URL")
+                .replace("ssl", "SSL")
+                .replace("api", "API")
+            )
+
+        for item in result:
+            row = []
+            for key in header_sorted:
+                entry = item.get(key)
+                row.append(str(entry) if entry is not None else "-")
+
+            table.add_row(*row)
+
+        console = Console()
+        console.print(table, justify="center")
+
+    sys.exit(0)
+
+
+@cli.command("add")
+@click.pass_context
+@click.argument("requirement", nargs=-1, metavar="PKG")
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment to which the given package should be added.",
+)
+@click.option(
+    "--index-url",
+    "-i",
+    default="https://pypi.org/simple",
+    type=str,
+    metavar="INDEX_URL",
+    show_default=True,
+    help="Specify Python package index to be used as a source for the given requirement.",
+)
+@click.option(
+    "--dev",
+    is_flag=True,
+    show_default=True,
+    help="Add the given package to the development packages.",
+)
+@handle_cli_exception
+def add(
+    requirement: typing.List[str],
+    runtime_environment: typing.Optional[str],
+    index_url: str,
+    dev: bool,
+) -> None:
+    """Add one or multiple requirements to the project.
+
+    Add one or multiple requirements to the direct dependency listing without actually installing them.
+    The supplied requirement is specified using PEP-508 standard.
+
+    Examples:
+      thamos add flask
+
+      thamos add tensorflow --runtime-environment "training"
+
+      thamos add --dev 'pytest~=6.2.0'
+
+      thamos add 'importlib-metadata; python_version < "3.8"'
+    """
+    project = configuration.get_project(runtime_environment, missing_dir_ok=True)
+    for req in requirement:
+        _LOGGER.info(
+            "Adding %r to %s requirements of runtime environment %r",
+            req,
+            "development" if dev else "default",
+            project.runtime_environment.name,
+        )
+        project.pipfile.add_requirement(
+            req, is_dev=dev, index_url=index_url, force=True
+        )
+
+    _LOGGER.warning(
+        "Changes done might require triggering new advise to resolve dependencies"
+    )
+    configuration.save_project(project)
+
+
+@cli.command("remove")
+@click.pass_context
+@click.argument("requirement", nargs=-1, metavar="PKG")
+@click.option(
+    "--runtime-environment",
+    "-r",
+    default=None,
+    metavar="NAME",
+    envvar="THAMOS_RUNTIME_ENVIRONMENT",
+    help="Specify runtime environment from which the given package should be removed.",
+)
+@handle_cli_exception
+def remove(
+    requirement: typing.List[str],
+    runtime_environment: typing.Optional[str],
+) -> None:
+    """Remove the given requirement.
+
+    Examples:
+      thamos remove flask
+
+      thamos remove pytest --runtime-environment "training"
+    """
+    project = configuration.get_project(runtime_environment)
+
+    for req in requirement:
+        any_change = False
+        if req in project.pipfile.packages.packages:
+            project.pipfile.packages.packages.pop(req)
+            _LOGGER.info(
+                "Removed %r from default requirements for runtime environment %r",
+                req,
+                project.runtime_environment.name,
+            )
+            any_change = True
+
+        if req in project.pipfile.dev_packages.packages:
+            project.pipfile.dev_packages.packages.pop(req)
+            _LOGGER.info(
+                "Removed %r from development requirements for runtime environment %r",
+                req,
+                project.runtime_environment.name,
+            )
+            any_change = True
+
+        if not any_change:
+            _LOGGER.error(
+                "Requirement %r not found in project requirements for runtime environment %r, "
+                "aborting making any changes",
+                req,
+                project.runtime_environment.name,
+            )
+            sys.exit(1)
+
+    _LOGGER.warning(
+        "Changes done might require triggering new advise to resolve dependencies"
+    )
+    configuration.save_project(project)
+
+
 __name__ == "__main__" and cli()
