@@ -19,7 +19,11 @@
 
 import logging
 import os
-import typing
+import sys
+from typing import Any
+from typing import Dict
+from typing import Optional
+from typing import List
 from urllib.parse import urljoin
 from jsonschema import validate
 
@@ -27,8 +31,12 @@ import click
 import requests
 import yaml
 
-from thoth.common import normalize_os_version
 from thoth.common import map_os_name
+from thoth.common import normalize_os_version
+from thoth.common import RuntimeEnvironment
+from thoth.python import Pipfile
+from thoth.python import Project
+from virtualenv import cli_run as virtualenv_cli_run
 
 from .utils import workdir
 from .discover import discover_cpu
@@ -36,8 +44,10 @@ from .discover import discover_cuda_version
 from .discover import discover_distribution
 from .discover import discover_python_version
 from .discover import discover_platform
+from .discover import discover_base_image
 from .exceptions import NoApiSupported
 from .exceptions import NoRuntimeEnvironmentError
+from .exceptions import RuntimeEnvironmentExistsError
 from .exceptions import ConfigurationError
 from .exceptions import NoProjectDirError
 from .exceptions import ServiceUnavailable
@@ -48,10 +58,59 @@ _API_CONNECTION_TIMEOUT = int(os.getenv("THAMOS_API_CONNECTION_TIMEOUT", 5))
 
 # The schema is enforcing all the options. This will make sure the right version of Thamos is
 # installed and no configuration options are silently ignored.
+_CONFIG_RUNTIME_ENVIRONMENT_SCHEMA = {
+    "type": "object",
+    "properties": {
+        "cuda_version": {"type": ["string", "null"]},
+        "openblas_version": {"type": ["string", "null"]},
+        "openmpi_version": {"type": ["string", "null"]},
+        "cudnn_version": {"type": ["string", "null"]},
+        "mkl_version": {"type": ["string", "null"]},
+        "base_image": {"type": ["string", "null"]},
+        "hardware": {
+            "type": "object",
+            "properties": {
+                "cpu_family": {"type": "integer"},
+                "cpu_model": {"type": "integer"},
+                "gpu_model": {"type": ["string", "null"]},
+            },
+            "required": [],
+            "additionalProperties": False,
+        },
+        "name": {"type": "string", "pattern": r"^[a-zA-Z0-9_-]+$"},
+        "operating_system": {
+            "type": "object",
+            "properties": {
+                "name": {"type": ["string", "null"]},
+                "version": {"type": ["string", "null"]},
+            },
+            "required": ["name", "version"],
+            "additionalProperties": False,
+        },
+        "platform": {"type": ["string", "null"]},
+        "python_version": {
+            "type": "string",
+            "pattern": r"^[0-9]+\.[0-9]+$",
+        },
+        "recommendation_type": {
+            "type": "string",
+            "enum": [
+                "latest",
+                "stable",
+                "performance",
+                "security",
+                "testing",
+            ],
+        },
+    },
+    "additionalProperties": False,
+    "required": ["name"],
+}
 _CONFIG_SCHEMA = {
     "type": "object",
     "properties": {
         "host": {"type": "string", "format": "hostname"},
+        "virtualenv": {"type": "boolean"},
         "tls_verify": {"type": "boolean"},
         "requirements_format": {
             "type": "string",
@@ -59,48 +118,7 @@ _CONFIG_SCHEMA = {
         },
         "runtime_environments": {
             "type": "array",
-            "items": {
-                "type": "object",
-                "properties": {
-                    "cuda_version": {"type": "string"},
-                    "hardware": {
-                        "type": "object",
-                        "properties": {
-                            "cpu_family": {"type": "integer"},
-                            "cpu_model": {"type": "integer"},
-                        },
-                        "required": [],
-                        "additionalProperties": False,
-                    },
-                    "name": {"type": "string", "pattern": r"^[a-zA-Z0-9:_-]+$"},
-                    "operating_system": {
-                        "type": "object",
-                        "properties": {
-                            "name": {"type": "string"},
-                            "version": {"type": "string"},
-                        },
-                        "required": ["name", "version"],
-                        "additionalProperties": False,
-                    },
-                    "platform": {"type": "string"},
-                    "python_version": {
-                        "type": "string",
-                        "pattern": r"^[0-9]+\.[0-9]+$",
-                    },
-                    "recommendation_type": {
-                        "type": "string",
-                        "enum": [
-                            "latest",
-                            "stable",
-                            "performance",
-                            "security",
-                            "testing",
-                        ],
-                    },
-                },
-                "additionalProperties": False,
-                "required": ["name"],
-            },
+            "items": _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA,
         },
         "managers": {
             "type": "array",
@@ -137,6 +155,7 @@ class _Configuration:
         self.explicit_host = None
         self.tls_verify = None
         self._api_url = None
+        self.config_path = None
 
     @property
     def api_url(self):
@@ -182,6 +201,42 @@ class _Configuration:
         response.raise_for_status()
         return response.headers.get("X-Thoth-Version", "Not Available")
 
+    def get_virtualenv_path(
+        self, runtime_environment: Optional[str] = None
+    ) -> Optional[str]:
+        """Get path to a virtual environment."""
+        if not self.content.get("virtualenv", False):
+            return None
+
+        if sys.base_prefix != sys.prefix:
+            _LOGGER.warning(
+                "Detected running in a virtual environment, no managed virtual environment will be used"
+            )
+            return None
+
+        return os.path.join(self.get_overlays_directory(runtime_environment), ".venv")
+
+    def create_virtualenv(self, runtime_environment: Optional[str] = None) -> None:
+        """Create a virtual environment for the given runtime environment."""
+        virtualenv_path = self.get_virtualenv_path(runtime_environment)
+        if virtualenv_path is None:
+            raise ConfigurationError("No virtual environment configured")
+
+        virtualenv_args = [virtualenv_path]
+        python_version = self.get_runtime_environment(runtime_environment).get(
+            "python_version"
+        )
+        if python_version:
+            virtualenv_args.extend(["--python", python_version])
+
+        _LOGGER.info("Creating virtual environment")
+        _LOGGER.debug(
+            "Virtual environment will be created in %r using %r",
+            virtualenv_path,
+            virtualenv_args,
+        )
+        virtualenv_cli_run(virtualenv_args)  # Raises on any error.
+
     def config_file_exists(self) -> bool:
         """Check if configuration file exists."""
         try:
@@ -211,15 +266,35 @@ class _Configuration:
 
             self._configuration = yaml.safe_load(self._configuration)
 
+    def reset_config(self) -> None:
+        """Discard loaded config in memory."""
+        self._configuration = None
+        self.config_path = None
+
     def load_config(self, force: bool = False) -> None:
         """Load configuration from a file."""
         if not self._configuration and not force:
             with workdir(config.CONFIG_NAME):
+                self.config_path = os.getcwd()
                 self.load_config_from_file(config.CONFIG_NAME)
 
+    def save_config(self, path: Optional[str] = None) -> None:
+        """Save the configuration to disc."""
+        if path:
+            with open(path, "w") as f:
+                yaml.dump(self.content, f)
+            _LOGGER.debug("Configuration changes written to %r", path)
+        else:
+            with workdir(config.CONFIG_NAME), open(config.CONFIG_NAME, "w") as f:
+                yaml.dump(self.content, f)
+                _LOGGER.debug(
+                    "Configuration changes written to %r",
+                    os.path.join(os.getcwd(), config.CONFIG_NAME),
+                )
+
     def create_default_config(
         self, template: str = None, nowrite: bool = False
-    ) -> typing.Optional[dict]:
+    ) -> Optional[dict]:
         """Place default configuration into the current directory."""
         if not os.path.isdir(".git"):
             _LOGGER.warning("Configuration file is not created in the root of git repo")
@@ -234,12 +309,14 @@ class _Configuration:
         cpu_info = discover_cpu()
         cuda_version = discover_cuda_version()
         # Add quotes for textual representation in the config file.
-        cuda_version = f"'{cuda_version}" if cuda_version is not None else "null"
+        cuda_version = f"'{cuda_version}'" if cuda_version is not None else "null"
         os_name, os_version = discover_distribution()
         os_name = map_os_name(os_name)
         os_version = normalize_os_version(os_name, os_version)
         python_version = discover_python_version()
         platform = discover_platform()
+        base_image = discover_base_image()
+        base_image = base_image if base_image is not None else "null"
 
         requirements_format = os.getenv(
             "THAMOS_REQUIREMENTS_FORMAT", self._DEFAULT_REQUIREMENTS_FORMAT
@@ -261,6 +338,7 @@ class _Configuration:
             python_version=python_version,
             platform=platform,
             requirements_format=requirements_format,
+            base_image=base_image,
             **cpu_info,
             **(dict(os.environ) if expand_env else {}),
         )
@@ -288,7 +366,32 @@ class _Configuration:
         """List available runtime environments."""
         return self.content.get("runtime_environments", [])
 
-    def get_runtime_environment(self, name: str = None) -> typing.Optional[dict]:
+    def set_runtime_environment(
+        self, runtime_environment: Dict[str, Any], force: bool = False
+    ) -> None:
+        """Add a runtime environment entry, overrides already existing one if force was set."""
+        try:
+            validate(runtime_environment, _CONFIG_RUNTIME_ENVIRONMENT_SCHEMA)
+        except Exception:
+            _LOGGER.error("Failed to validate runtime environment entry against schema")
+            raise
+
+        try:
+            existing = self.get_runtime_environment(runtime_environment["name"])
+        except NoRuntimeEnvironmentError:
+            self.content.setdefault("runtime_environments", []).append(
+                runtime_environment
+            )
+        else:
+            if force:
+                existing.clear()
+                existing.update(runtime_environment)
+            else:
+                raise RuntimeEnvironmentExistsError(
+                    f"Runtime environment {runtime_environment['name']!r} already exists"
+                )
+
+    def get_runtime_environment(self, name: Optional[str] = None) -> Optional[dict]:
         """Get runtime environment, retrieve the first runtime environment (the default one) if no name is provided."""
         content = self.content
         if "runtime_environments" not in content:
@@ -389,7 +492,7 @@ class _Configuration:
 
     def check_runtime_environment(
         self, runtime_environment_name: str
-    ) -> typing.List[typing.Dict[str, typing.Any]]:
+    ) -> List[Dict[str, Any]]:
         """Check the given runtime environment entry."""
         runtime_environment = self.get_runtime_environment(runtime_environment_name)
 
@@ -498,8 +601,8 @@ class _Configuration:
         return result
 
     def check(
-        self, runtime_environment_name: typing.Optional[str] = None
-    ) -> typing.List[typing.Dict[str, typing.Any]]:
+        self, runtime_environment_name: Optional[str] = None
+    ) -> List[Dict[str, Any]]:
         """Check the configuration file and produce a report."""
         result = []
 
@@ -517,6 +620,23 @@ class _Configuration:
                 }
             )
 
+        if self.content.get("overlays_dir"):
+            with workdir():
+                for file_type in (
+                    "Pipfile",
+                    "Pipfile.lock",
+                    "requirements.txt",
+                    "requirements.in",
+                ):
+                    if os.path.isfile(file_type):
+                        result.append(
+                            {
+                                "message": f"Overlays configured but {file_type!r} file found in the repo root, "
+                                f"this might lead to misleading repository interpretation",
+                                "type": "WARNING",
+                            }
+                        )
+
         if runtime_environment_name is not None:
             result.extend(self.check_runtime_environment(runtime_environment_name))
         else:
@@ -537,7 +657,10 @@ class _Configuration:
             raise
 
     def get_overlays_directory(
-        self, runtime_environment_name: typing.Optional[str] = None
+        self,
+        runtime_environment_name: Optional[str] = None,
+        *,
+        missing_dir_ok: bool = False,
     ) -> str:
         """Get path to an overlays directory."""
         runtime_environment_config = self.get_runtime_environment(
@@ -550,7 +673,106 @@ class _Configuration:
             if overlays_dir is None:
                 return os.getcwd()
 
-            return os.path.join(overlays_dir, runtime_environment_config["name"])
+            path = os.path.join(overlays_dir, runtime_environment_config["name"])
+            if not missing_dir_ok and not os.path.isdir(path):
+                suffix = (
+                    f" --runtime-environment {runtime_environment_config['name']!r}"
+                    if runtime_environment_name
+                    else ""
+                )
+                raise ConfigurationError(
+                    f"The directory structure for {runtime_environment_config['name']!r} is not initialized yet, "
+                    f"you can initialize it by adding packages using "
+                    f"`thamos add <pkg>{suffix}`",
+                )
+
+            return os.path.abspath(path)
+
+    def get_project(
+        self,
+        runtime_environment_name: Optional[str] = None,
+        *,
+        missing_dir_ok: bool = False,
+    ) -> Project:
+        """Get the given overlay."""
+        path = self.get_overlays_directory(
+            runtime_environment_name=runtime_environment_name,
+            missing_dir_ok=missing_dir_ok,
+        )
+        runtime_environment = RuntimeEnvironment.from_dict(
+            self.get_runtime_environment(runtime_environment_name)
+        )
+        if self.requirements_format == "pipenv":
+            pipfile_lock_path = os.path.join(path, "Pipfile.lock")
+            if not os.path.exists(pipfile_lock_path):
+                pipfile_lock_path = None
+
+            pipfile_path = os.path.join(path, "Pipfile")
+            if not os.path.isfile(pipfile_path):
+                if not os.path.isdir(path):
+                    _LOGGER.info("Creating directory structure in %r", path)
+                    os.makedirs(path, exist_ok=True)
+                pipfile = Pipfile.from_dict({})
+                pipfile.to_file(path=pipfile_path)
+
+            project = Project.from_files(
+                pipfile_path=pipfile_path,
+                pipfile_lock_path=pipfile_lock_path,
+                runtime_environment=runtime_environment,
+                without_pipfile_lock=pipfile_lock_path is None,
+            )
+        else:
+            requirements_in_file_path = os.path.join(path, "requirements.in")
+            if not os.path.isfile(requirements_in_file_path):
+                requirements_txt_file_path = os.path.join(path, "requirements.txt")
+                if os.path.isfile(requirements_txt_file_path):
+                    _LOGGER.warning(
+                        "Using %r for direct dependencies", requirements_in_file_path
+                    )
+                    project = Project.from_pip_compile_files(
+                        requirements_path=requirements_txt_file_path,
+                        requirements_lock_path=None,
+                        allow_without_lock=True,
+                        runtime_environment=runtime_environment,
+                    )
+                else:
+                    raise NotImplementedError(
+                        "No requirements.txt/requirements.in files found, it is recommended to "
+                        "use Pipenv files for managing dependencies"
+                    )
+            else:
+                project = Project.from_pip_compile_files(
+                    requirements_path=requirements_in_file_path,
+                    requirements_lock_path=None,
+                    allow_without_lock=True,
+                    runtime_environment=runtime_environment,
+                )
+
+        return project
+
+    def save_project(self, project: Project) -> None:
+        """Save the given project to disc, performs noop if project is not dirty."""
+        old_project = self.get_project(project.runtime_environment.name)
+        if old_project.runtime_environment != project.runtime_environment:
+            self.set_runtime_environment(
+                runtime_environment=project.runtime_environment
+            )
+            self.save_config()
+
+        if old_project.pipfile != project.pipfile:
+            if config.requirements_format == "pipenv":
+                pipfile_path = os.path.join(
+                    self.get_overlays_directory(project.runtime_environment.name)
+                )
+                project.pipfile.to_file(path=pipfile_path, keep_thoth_section=True)
+            else:
+                requirements_in_file = self.get_overlays_directory(
+                    project.runtime_environment.name
+                )
+                with open(
+                    os.path.join(requirements_in_file, "requirements.in"), "w"
+                ) as f:
+                    f.write(project.pipfile.to_requirements_file(develop=False))
 
 
 config = _Configuration()
