@@ -48,16 +48,19 @@ from .swagger_client import ApiClient
 from .swagger_client import BuildAnalysisApi
 from .swagger_client import Configuration
 from .swagger_client import PythonStack
-from .swagger_client import RuntimeEnvironment
 from .swagger_client import AdviseInput
 from .swagger_client import AdviseApi
 from .swagger_client import ImageAnalysisApi
 from .swagger_client import ProvenanceApi
+from .swagger_client import HardwareApi
+from .swagger_client import S2iApi
+from .swagger_client import PythonPackagesApi
 from .swagger_client.models import AnalysisResultResponse
 from .config import config as thoth_config
 from .exceptions import UnknownAnalysisType
 from .exceptions import TimeoutError
 from .exceptions import ApiError
+from .exceptions import NoDevRequirements
 from .exceptions import NoRequirementsFile
 
 from typing import Callable, Any, Union, Dict
@@ -113,7 +116,7 @@ def is_analysis_ready(analysis_id: str) -> bool:
     if not host:
         thoth_config.load_config()
         host = thoth_config.content.get("host") or config.host
-    source = analysis_id.rsplit("-", 1)[0]
+    source = analysis_id.rsplit("-", maxsplit=2)[0]
     source_url = _SOURCE.get(source)
     response = requests.get(f"https://{host}/api/v1/{source_url}/{analysis_id}")
     if response.status_code == 202:
@@ -123,11 +126,12 @@ def is_analysis_ready(analysis_id: str) -> bool:
         return True
 
     raise ApiError(
-        f"Thoth Backend didn't respond with correct status code. Returned code - {response.status_code}"
+        f"Thoth Backend didn't respond with correct status code. Returned code - "
+        f"{response.status_code}: {response.text}"
     )
 
 
-def _wait_for_analysis(status_func: Callable[..., Any], analysis_id: str) -> None:
+def _wait_for_analysis(status_func: Callable[..., Any], analysis_id: str, timeout: typing.Optional[int] = None) -> None:
     """Wait for ongoing analysis to finish."""
     # noqa
     @contextmanager
@@ -147,10 +151,11 @@ def _wait_for_analysis(status_func: Callable[..., Any], analysis_id: str) -> Non
 
     sleep_time = 0.5
     retries = 0
+    timeout = timeout if timeout is not None else _THAMOS_TIMEOUT
     with spinner():
         start_time = monotonic()
         while True:
-            if _THAMOS_TIMEOUT and monotonic() - start_time > _THAMOS_TIMEOUT:
+            if timeout and monotonic() - start_time > timeout:
                 raise TimeoutError(
                     f"Thoth backend did not respond in time, timeout set "
                     f"to {_THAMOS_TIMEOUT} - see {jl('thamos_timeout')}"
@@ -252,13 +257,13 @@ def _retrieve_analysis_result(
             sleep(_RETRY_ON_ERROR_SLEEP)
 
 
-def _get_static_analysis() -> typing.Optional[dict]:
+def _get_static_analysis(src_path: str = ".") -> typing.Optional[dict]:
     """Get static analysis of files used in project."""
     # We are running in the root directory of project, use the root part for gathering static analysis.
     _LOGGER.info("Performing static analysis of sources to gather library usage")
     try:
         library_usage = gather_library_usage(
-            ".", ignore_errors=True, without_standard_imports=True
+            src_path, ignore_errors=True, without_standard_imports=True
         )
     except FileNotFoundError:
         _LOGGER.warning("No library usage was aggregated - no Python sources found")
@@ -307,17 +312,19 @@ def advise_using_config(
     pipfile_lock: str,
     config: str = None,
     *,
-    runtime_environment_name: str = None,
+    runtime_environment_name: typing.Optional[str] = None,
+    src_path: typing.Optional[str] = None,
     recommendation_type: str = None,
     dev: bool = False,
     no_static_analysis: bool = False,
     no_user_stack: bool = False,
     nowait: bool = False,
     force: bool = False,
-    limit: int = None,
+    limit: typing.Optional[int] = None,
     count: int = 1,
     debug: bool = False,
-    origin: str = None,
+    origin: typing.Optional[str] = None,
+    timeout: typing.Optional[int] = None,
     github_event_type: typing.Optional[str] = None,
     github_check_run_id: typing.Optional[int] = None,
     github_installation_id: typing.Optional[int] = None,
@@ -334,6 +341,7 @@ def advise_using_config(
         pipfile=pipfile,
         pipfile_lock=pipfile_lock,
         recommendation_type=recommendation_type,
+        src_path=src_path if src_path is not None else thoth_config.config_path,
         runtime_environment=thoth_config.get_runtime_environment(
             runtime_environment_name
         ),
@@ -347,6 +355,7 @@ def advise_using_config(
         count=count,
         debug=debug,
         origin=origin,
+        timout=timeout,
         github_event_type=github_event_type,
         github_check_run_id=github_check_run_id,
         github_installation_id=github_installation_id,
@@ -363,6 +372,7 @@ def advise(
     recommendation_type: str = None,
     *,
     runtime_environment: dict = None,
+    src_path: str = ".",
     runtime_environment_name: str = None,
     dev: bool = False,
     no_static_analysis: bool = False,
@@ -372,12 +382,14 @@ def advise(
     limit: int = None,
     count: int = 1,
     debug: bool = False,
-    origin: str = None,
+    origin: typing.Optional[str] = None,
+    timeout: typing.Optional[int] = None,
     github_event_type: typing.Optional[str] = None,
     github_check_run_id: typing.Optional[int] = None,
     github_installation_id: typing.Optional[int] = None,
     github_base_repo_url: typing.Optional[str] = None,
     source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
+    kebechet_metadata: typing.Optional[Dict] = None,
 ) -> typing.Optional[tuple]:
     """Submit a stack for adviser checks and wait for results."""
     if not pipfile:
@@ -414,7 +426,7 @@ def advise(
 
     library_usage = None
     if not no_static_analysis:
-        library_usage = _get_static_analysis()
+        library_usage = _get_static_analysis(src_path)
         _LOGGER.debug(
             "Library usage:%s",
             "\n" + json.dumps(library_usage, indent=2) if library_usage else None,
@@ -426,8 +438,6 @@ def advise(
         # Override recommendation type specified explicitly in the runtime environment entry.
         runtime_environment.pop("recommendation_type", None)
 
-        runtime_environment = RuntimeEnvironment(**runtime_environment)  # type: ignore
-
     advise_input = AdviseInput(
         stack, runtime_environment=runtime_environment, library_usage=library_usage
     )
@@ -469,6 +479,9 @@ def advise(
     if github_base_repo_url is not None:
         parameters["github_base_repo_url"] = github_base_repo_url
 
+    if kebechet_metadata is not None:
+        parameters["kebechet_metadata"] = kebechet_metadata
+
     response = api_instance.post_advise_python(advise_input, **parameters)
 
     _LOGGER.info(
@@ -485,7 +498,7 @@ def advise(
     if nowait:
         return response.analysis_id
     # We call custom status function for advise until swagger client supports mulitple response codes.
-    _wait_for_analysis(is_analysis_ready, response.analysis_id)
+    _wait_for_analysis(is_analysis_ready, response.analysis_id, timeout)
     _LOGGER.debug("Retrieving adviser result for %r", response.analysis_id)
     response = _retrieve_analysis_result(
         api_instance.get_advise_python, response.analysis_id
@@ -502,6 +515,7 @@ def advise_here(
     recommendation_type: typing.Optional[str] = None,
     *,
     runtime_environment: dict = None,
+    src_path: str = ".",
     runtime_environment_name: typing.Optional[str] = None,
     dev: bool = False,
     no_static_analysis: bool = False,
@@ -511,12 +525,14 @@ def advise_here(
     limit: typing.Optional[int] = None,
     count: int = 1,
     debug: bool = False,
+    timeout: typing.Optional[int] = None,
     origin: typing.Optional[str] = None,
     github_event_type: typing.Optional[str] = None,
     github_check_run_id: typing.Optional[int] = None,
     github_installation_id: typing.Optional[int] = None,
     github_base_repo_url: typing.Optional[str] = None,
     source_type: typing.Optional[ThothAdviserIntegrationEnum] = None,
+    kebechet_metadata: typing.Optional[Dict] = None,
 ) -> typing.Optional[tuple]:
     """Run advise in current directory, requires no arguments."""
     requirements_format = thoth_config.requirements_format
@@ -565,6 +581,7 @@ def advise_here(
         pipfile=pipfile,
         pipfile_lock=pipfile_lock_str,
         recommendation_type=recommendation_type,
+        src_path=src_path,
         runtime_environment=runtime_environment,
         runtime_environment_name=runtime_environment_name,
         dev=dev,
@@ -576,11 +593,13 @@ def advise_here(
         count=count,
         debug=debug,
         origin=origin,
+        timeout=timeout,
         source_type=source_type,
         github_event_type=github_event_type,
         github_check_run_id=github_check_run_id,
         github_installation_id=github_installation_id,
         github_base_repo_url=github_base_repo_url,
+        kebechet_metadata=kebechet_metadata,
     )
 
 
@@ -593,7 +612,8 @@ def provenance_check(
     nowait: bool = False,
     force: bool = False,
     debug: bool = False,
-    origin: str = None,
+    origin: typing.Optional[str] = None,
+    timeout: typing.Optional[int] = None,
 ) -> typing.Optional[tuple]:
     """Submit a stack for provenance checks and wait for results."""
     if not pipfile:
@@ -615,7 +635,7 @@ def provenance_check(
     if nowait:
         return response.analysis_id
 
-    _wait_for_analysis(is_analysis_ready, response.analysis_id)
+    _wait_for_analysis(is_analysis_ready, response.analysis_id, timeout)
     _LOGGER.debug("Retrieving provenance check result for %r", response.analysis_id)
     response = _retrieve_analysis_result(
         api_instance.get_provenance_python, response.analysis_id
@@ -632,7 +652,8 @@ def provenance_check_here(
     nowait: bool = False,
     force: bool = False,
     debug: bool = False,
-    origin: str = None,
+    origin: typing.Optional[str] = None,
+    timeout: typing.Optional[int] = None,
 ) -> typing.Optional[tuple]:
     """Submit a provenance check in current directory."""
     if not os.path.isfile("Pipfile"):
@@ -649,6 +670,7 @@ def provenance_check_here(
             force=force,
             debug=debug,
             origin=origin,
+            timeout=timeout,
         )
 
 
@@ -663,6 +685,7 @@ def image_analysis(
     verify_tls: bool = True,
     nowait: bool = False,
     force: bool = False,
+    timeout: typing.Optional[int] = None,
     debug: bool = False,
 ) -> Union[Dict, str, None]:
     """Submit an image for analysis to Thoth."""
@@ -695,7 +718,7 @@ def image_analysis(
     if nowait:
         return response.analysis_id
 
-    _wait_for_analysis(is_analysis_ready, response.analysis_id)
+    _wait_for_analysis(is_analysis_ready, response.analysis_id, timeout)
     _LOGGER.debug(
         "Retrieving image analysis result result for %r", response.analysis_id
     )
@@ -901,7 +924,9 @@ def install_using_config(
 
 
 def install(
-    runtime_environment_name: typing.Optional[str] = None, dev: bool = False
+    runtime_environment_name: typing.Optional[str] = None,
+    dev: bool = False,
+    pip_args: typing.Optional[typing.Tuple[str]] = None,
 ) -> None:
     """Perform installation of packages for the given runtime environment.
 
@@ -920,21 +945,96 @@ def install(
         if method == "pipenv":
             if not os.path.isfile("Pipfile.lock"):
                 raise NoRequirementsFile(
-                    f"No Pipfile.lock found in {os.getcwd()!r} needed to install dependencies"
+                    f"No Pipfile.lock found in {os.getcwd()!r} needed to install dependencies, "
+                    "issue `thamos advise` resolve dpeendencies"
                 )
             if not os.path.isfile("Pipfile"):  # Required for computing digests.
                 raise NoRequirementsFile(
                     f"No Pipfile found in {os.getcwd()!r} needed for the installation process"
                 )
+
+            if dev:
+                with open("Pipfile.lock") as pipfile_lock_file:
+                    content = json.load(pipfile_lock_file)
+
+                if not content.get("develop"):
+                    raise NoDevRequirements(
+                        "No development requirements found in the lock file, make sure development "
+                        "requirements are stated and the resolved stack preserves them by "
+                        "using `thamos advise --dev`"
+                    )
+
+                del content
         else:
             if not os.path.isfile("requirements.txt"):
                 raise NoRequirementsFile(
                     f"No requirements.txt file found in {os.getcwd()!r} needed to install dependencies"
                 )
 
+        if runtime_environment_name is None:
+            config_entry = thoth_config.get_runtime_environment(runtime_environment_name)
+            runtime_environment_name = config_entry["name"]
+        _LOGGER.info("Installing requirements for runtime environment %r", runtime_environment_name)
+
         _LOGGER.info(
             "Using %r installation method to install dependencies stated in %r",
             method,
             os.getcwd(),
         )
-        micropipenv.install(method=method, deploy=True, dev=dev)
+
+        micropipenv_kwargs = {
+            "method": method,
+            "deploy": True,
+            "dev": dev,
+            "pip_args": pip_args,
+        }
+
+        virtualenv_path = thoth_config.get_virtualenv_path(runtime_environment_name)
+        if virtualenv_path:
+            if not os.path.isdir(virtualenv_path):
+                thoth_config.create_virtualenv()
+
+            micropipenv_kwargs["pip_bin"] = os.path.join(virtualenv_path, "bin", "pip3")
+
+        # micropipenv writes and prints the lockfile which is not very user friendly when thamos is used by a user.
+        # Suppress this behavior unless these environment variables options are supplied explicitly.
+        old_write = os.getenv("MICROPIPENV_NO_LOCKFILE_WRITE")
+        if old_write is None:
+            os.environ["MICROPIPENV_NO_LOCKFILE_WRITE"] = "1"
+
+        old_print = os.getenv("MICROPIPENV_NO_LOCKFILE_PRINT")
+        if old_print is None:
+            os.environ["MICROPIPENV_NO_LOCKFILE_PRINT"] = "1"
+
+        try:
+            micropipenv.install(**micropipenv_kwargs)
+        finally:
+            if old_write is None:
+                os.environ.pop("MICROPIPENV_NO_LOCKFILE_WRITE", None)
+
+            if old_print is None:
+                os.environ.pop("MICROPIPENV_NO_LOCKFILE_PRINT", None)
+
+
+@with_api_client
+def list_hardware_environments(api_client: ApiClient) -> typing.Dict[str, Any]:
+    """Get information about hardware for which Thoth can give recommendations."""
+    return (
+        HardwareApi(api_client)
+        .list_hardware_environments()
+        .to_dict()["hardware_environments"]
+    )
+
+
+@with_api_client
+def list_thoth_s2i(api_client: ApiClient) -> typing.Dict[str, Any]:
+    """Get information about hardware for which Thoth can give recommendations."""
+    return S2iApi(api_client).list_s2i_python().to_dict()["s2i"]
+
+
+@with_api_client
+def list_python_package_indexes(api_client: ApiClient) -> typing.Dict[str, Any]:
+    """Get information about hardware for which Thoth can give recommendations."""
+    return (
+        PythonPackagesApi(api_client).list_python_package_indexes().to_dict()["indexes"]
+    )
